#!/usr/bin/perl
#
# $Id$
#
# Copyright (c) 2010, SIDN (Stichting Internet Domeinregistratie Nederland).
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions 
# are met:
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

use strict;
use warnings;

use POSIX qw(geteuid);
use Spreadsheet::ParseExcel;

use ActionWords;   # reserved action words

my $oExcel = new Spreadsheet::ParseExcel;

if (@ARGV != 2) {
  print "Too less arguments.\n" if (@ARGV < 2);
  print "Too many arguments.\n" if (@ARGV > 2);
  usage();
  exit;
}
die "You must provide a Excel workbook name and a worksheet name to $0.\n" unless @ARGV;

my $wkb = $oExcel->Parse($ARGV[0]);
my($row, $column, $wks, $cell);
my $argnum;
my %actionwords = (
  # Basic commands
  "RESET" => \&ActionWords::reset,
  "START" => \&ActionWords::start,
  "STOP" => \&ActionWords::stop,
  "WAIT" => \&ActionWords::wait,
  "COMMENT" => \&ActionWords::comment,
  # Preparation commands
  "LOADCONFIG" => \&ActionWords::loadConfig,
  "LOADKASP" => \&ActionWords::loadKASP,
  "LOADCONF" => \&ActionWords::loadConf,
  "INITREPO" => \&ActionWords::initRepo,
  "LOADZONE" => \&ActionWords::loadZone,
  "REMOVEZONE" => \&ActionWords::removeZone,
  "UPDATEALL" => \&ActionWords::updateAll,
  "BACKUPALL" => \&ActionWords::backupAll,
  "RETIREKSK" => \&ActionWords::retireKSK,
  "DSPUBLISHED" => \&ActionWords::DSPublished,
  "ROLLOVERKSK" => \&ActionWords::rolloverKSK,
  "ROLLOVERZSK" => \&ActionWords::rolloverZSK,
  # Check commands
  "COUNTRRSETS" => \&ActionWords::countRRSets,
  "CHECKKEYTAG" => \&ActionWords::checkKeytag,
  "CHECKLOG" => \&ActionWords::checkLog,
  "CHECKFILEPRESENT" => \&ActionWords::checkFilePresent,
  # Test cluster
  "TESTCASE" => \&ActionWords::testCase,
  "TESTCONDITION" => \&ActionWords::testCondition,
  # Other
  "&CONT" => \&ActionWords::actionwordContinue,
  "&RUNCOMMAND" => \&ActionWords::runCommand,
);
my %arguments = (
  "&CONT" => \&Arguments::argumentContinue,
  "&KEEP" => \&Arguments::keep,
  "&REFER" => \&Arguments::refer,
  "&EMPTY" => \&Arguments::empty,
  "&ANY" => \&Arguments::any,
  "&DATE" => \&Arguments::date,
);

sub main {
  my $row;
  my $col;
  my $line;
  my $actionword = "";
  my @arguments = ();
  my $continue = 0;

  # Program must be run with root privileges (root/sudo)
  if (geteuid() != 0) {
    print "Program can't be run as non-root.\nExiting.\n";
    exit(1);
  }

  print "FILE  : ", $wkb->{File}, "\n";
  print "COUNT : ", $wkb->{SheetCount}, "\n";

  print "AUTHOR: ", $wkb->{Author}, "\n"
    if defined $wkb->{Author};

  foreach $argnum (0 .. $#ARGV) {
    print "- Argument: $ARGV[$argnum]\n";
  }

  $wks = $wkb->Worksheet($ARGV[1]);
  print "SHEET: ", $wks->{Name}, "\n---------\n"
    if defined $wks->{Name};
  for(my $row = $wks->{MinRow}; defined $wks->{MaxRow} && $row <= $wks->{MaxRow}; $row++) {
    $continue = 0;
    # check for action word
    if (defined $wks->{Cells}[$row][0] && exists $actionwords{uc($wks->{Cells}[$row][0]->Value)}) {
      if ((uc $wks->{Cells}[$row][0]->Value eq '&CONT')) {
        # &cont found
      } else {
        # add row number to argument array for later use in actionword sub's
        $arguments[0] = $row+1;
        # get action word
        $actionword = $wks->{Cells}[$row][0]->Value;
        # get arguments
        for ($column = 1; $column <= $wks->{MaxCol}; $column++) {
          $cell = $wks->{Cells}[$row][$column];
          if (defined $cell && uc($cell->Value) eq '&CONT') {
            # next action word MUST be equal to &cont
            if (!ActionWords::checkForEqual(uc ($wks->{Cells}[$row+2][0]->Value), '&CONT')) {
              printf "%5.5d %s%d%s", $arguments[0], "ERROR: action word &cont NOT found in (", $arguments[0]+2, ",0). Program exit.\n";
              exit(1);
            }
            $continue = 1;
            $row++;   # continue on next row
            last;
          }
          # add parameter values to argument array
          if (defined $cell && uc($cell->Value) ne '&CONT' && length($cell->Value) > 0 && $cell->Value ne "") {
            $arguments[$column] = $cell->Value;
          }
        }
      }
    } else {
      # not a (defined) action word, print to report?
    }
    if (!$continue && $actionword ne '') {
      # call action word subroutine
      $actionwords{uc($actionword)}->(@arguments);

      # reset variables
      $actionword = "";
      $continue = 0;
      @arguments = ();
    }
  }
}

sub usage {
  print "usage: perl program.pl <workbook> <worksheet>\n";
  exit;
}

# ------------------------------------------------------------------------------
main();

# ------------------------------------------------------------------------------
# perl ActionWordParser.pl [workbook] [worksheet]+ [> [output file]]
#
# perl ActionWordParser.pl testwb.xls CBNM > output.txt
# ------------------------------------------------------------------------------
# $wks = $wkb->Worksheet('CBNM');
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
#		description	...	description
# actionword	arg[1]		...	arg[n]		&cont
#		description	...	description
# &cont		arg[n+1]	...	arg[n+m]
# ------------------------------------------------------------------------------
