#
# $Id$
#
# Copyright (c) 2010, SIDN (Stichting Internet Domeinregistratie Nederland).
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions 
# are met:
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

package ActionWords;

use Time::HiRes qw(gettimeofday);
use IPC::Open3;

my $testcondition = "";

# - reserved action words ------------------------------------------------------

# ------------------------------------------------------------------------------
# NOTE: Check commands (CountRRSets, CheckKeytag, CheckLog, CheckFilePresent)
#       can fail but don't abort the test.
#       Basic commands (Reset, Start, Stop) don't fail and can't abort the test.
#       Preparation commands (LoadConfig, LoadKASP, LoadConf, LoadZone,
#       RemoveZone, UpdateAll, BackupAll, RolloverKSK, RolloverZSK) don't fail
#       and can abort the test.
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
# Action word: Reset
#
# Arguments:
#
# Description: Resets the initial test situation. For that it runs the following
#              commands:
#              `rm -rf /var/opendnssec/tmp/* /var/opendnssec/signed/* /var/opendnssec/signconf/*`
#              `cp /etc/opendnssec/kasp.xml.sample /etc/opendnssec/kasp.xml`
#              `cp /etc/opendnssec/conf.xml.sample /etc/opendnssec/conf.xml`
#              `cp /etc/opendnssec/zonelist.xml.sample /etc/opendnssec/zonelist.xml`
#              `ods-ksmutil setup`
#
# Note: Bash - $? reads the exit status of the last command executed. After a 
#              function returns, $? gives the exit status of the last command
#              executed in the function.
# ------------------------------------------------------------------------------
sub reset {
  my @rc = ();

  printActionWordHeader($_[0], "Reset()");

  @rc = runCommand('rm -rf /var/opendnssec/tmp/* /var/opendnssec/signed/* /var/opendnssec/signconf/*');
  checkReturnCode($rc[2], 0);
  @rc = runCommand('cp /etc/opendnssec/kasp.xml.sample /etc/opendnssec/kasp.xml');
  checkReturnCode($rc[2], 0);
  @rc = runCommand('cp /etc/opendnssec/conf.xml.sample /etc/opendnssec/conf.xml');
  checkReturnCode($rc[2], 0);
  @rc = runCommand('cp /etc/opendnssec/zonelist.xml.sample /etc/opendnssec/zonelist.xml');
  checkReturnCode($rc[2], 0);
  @rc = runCommand('echo \'y\' | ods-ksmutil setup');
  checkReturnCode($rc[2], 0);
}


# ------------------------------------------------------------------------------
# Action word: start
#
# Arguments:
#
# Description: Basicly runs "ods-control start" and check if the deamons started
#              by running "ps -e | grep enforcer" and "ps -e | grep signer".
#              The command "ods-control start" starts two processes, 
#               1) /usr/local/sbin/ods-signer start
#               2) /usr/local/sbin/ods-enforcerd
# ------------------------------------------------------------------------------
sub start {
  my @rc = ();

  printActionWordHeader($_[0], "Start()");

  @rc = runCommand('ods-control start', 1);
  checkReturnCode($rc[2], 0);
  @rc = runCommand('ps -e | awk \'{print $4}\' | grep -x ods-enforcerd');
  checkReturnCode($rc[2], 0);
  @rc = runCommand('ps -ef | grep signer | grep -v grep | awk \'{print $8 " " $9}\'');
  checkReturnCode($rc[2], 0);
}


# ------------------------------------------------------------------------------
# Action word: stop
#
# Arguments:
#
# Description: Basicly runs "ods-control stop" and check if it stopped by
#              running "ps -e | grep enforcer" and "ps -e | grep signer" which
#              should show no running processes, else run "kill" command on
#              these processes.
# ------------------------------------------------------------------------------
sub stop {
  my @rc = ();

  printActionWordHeader($_[0], "Stop()");

  @rc = runCommand('ods-control stop');
  checkReturnCode($rc[2], 0);
  @rc = runCommand('ps -e | grep enforcer');
  checkReturnCode($rc[2], 0);   # if 0 then runCommand('kill ...')
  @rc = runCommand('ps -e | grep signer');
  @rc = checkReturnCode($rc[2], 0);   # if 0 then runCommand('kill ...')
}


# ------------------------------------------------------------------------------
# Action word: wait
#
# Arguments:   seconds
#
# Description: Causes the test script to sleep for the number of seconds, or
#              forever if no seconds.
#              Returns the number of seconds actually slept.
# ------------------------------------------------------------------------------
sub wait {
  my ($row, $sec) = @_;

  printActionWordHeader($row, "Wait($sec)\n");

  return sleep($sec);
}


# ------------------------------------------------------------------------------
# Action word: Comment
#
# Arguments:   text
#
# Description: Comment line is printed to output for logging purposes.
# ------------------------------------------------------------------------------
sub comment {
  my ($row, $str) = @_;

  printActionWordHeader($row, "Comment: $str\n");
}


# ------------------------------------------------------------------------------
# Action word: RunCommand
#
# Arguments:   command, background process
#
# Description: Runs command. If background process == TRUE RunCommand doesn't
#              wait for rc.
# ------------------------------------------------------------------------------
sub runCommand {
  my ($cmd, $bg) = @_;
  my $pid;
  my $retval;
  my $out;
  my $err;
  my $ret;
  my @rc = ();

  print getTime() . " RunCommand($cmd)\n";

  $pid = open3(\*WRITER, \*READER, \*ERROR, $cmd) if defined $cmd;
  #if \*ERROR is 0, stderr goes to stdout

  if (!defined $bg) {
    while(my $output = <READER>) {
      chomp($output);
      $out .= $output . "\n";
    }
    chomp($out) if defined ($out);

    while(my $errout = <ERROR>) {
      chomp($errout);
      $err .= $errout . "\n";
    }
    chomp($err) if defined($err);

    print getTime() . " - runCommand: Process started, PID: $pid\n";
    waitpid($pid, 0) or die "$?\n";   # return value as $rc[2]
    $retval = $?;
    $retval = $retval >> 8;   # exit code of the called program
    chomp($retval);
    $ret .= $retval if defined ($retval);

    $rc[0] = $out;
    $rc[1] = $err;
    $rc[2] = $ret;

    printRunCommandOutput(@rc);

    return @rc;
  } else {
    print getTime( ) . " - runCommand: Background process started, PID: $pid\n";
    waitpid($pid, 0) or die "$?\n";
    $retval = $?;
    $retval = $retval >> 8;   # exit code of the called program
    chomp($retval);
    $ret .= $retval if defined ($retval);

    $rc[0] = "";
    $rc[1] = "";
    $rc[2] = $ret;

    printRunCommandOutput(@rc);

    return @rc;
  }
}

# ------------------------------------------------------------------------------
# Action word: LoadConfig
#
# Arguments:   type, file
#
# Description: Update config file, based on type, by the file found on given
#              path.
# ------------------------------------------------------------------------------
sub loadConfig {
  my ($row, $type, $file) = @_;
  my @rc = ();
  my $dir = "";

  printActionWordHeader($row, "LoadConfig()");

  if (uc $type eq 'KASP') {
    $dir = ' /etc/opendnssec/kasp.xml';
  } elsif (uc $type eq 'CONF') {
    $dir = ' /etc/opendnssec/conf.xml';
  } elsif (uc $type eq 'SOFTHSM') {
    $dir = ' /etc/softhsm.conf';
  } else {
    print getTime() . " ERROR:\tUnknown type, known types are: KASP, CONF, SOFTHSM\n";
    print getTime() . " ABORT:\tProgram exit\n";
    exit;
  }

  @rc= runCommand('cp ' . $file . $dir);
  checkReturnCode($rc[2], 0);

  unless (-e $dir) {
    print getTime() . " ERROR:\t$dir not found\n";
    print getTime() . " ABORT:\tProgram exit\n";
    exit;
  }
}

# ------------------------------------------------------------------------------
# Action word: LoadKASP
#
# Arguments:   file
#
# Description: Update current /etc/opendnssec/kasp.xml by the file found on
#              given path.
# ------------------------------------------------------------------------------
sub loadKASP {
  my ($row, $file) = @_;
  my @rc = ();

  printActionWordHeader($row, "LoadKASP()");

  @rc= runCommand('cp ' . $file . ' /etc/opendnssec/kasp.xml');
  checkReturnCode($rc[2], 0);

  unless (-e "/etc/opendnssec/kasp.xml") {
    print getTime() . " ERROR:\t/etc/opendnssec/kasp.xml not found\n";
    print getTime() . " ABORT:\tProgram exit\n";
    exit;
  }
}

# ------------------------------------------------------------------------------
# Action word: LoadConf
#
# Arguments:   file
#
# Description: Update current conf.xml by the file found on given path.
# ------------------------------------------------------------------------------
sub loadConf {
  my ($row, $file) = @_;
  my @rc = ();

  printActionWordHeader($row, "LoadConf()");

  @rc = runCommand('cp ' . $file . ' /etc/opendnssec/conf.xml');
  checkReturnCode($rc[2], 0);

  unless (-e "/etc/opendnssec/conf.xml") {
    print getTime() . " ERROR:\t/etc/opendnssec/conf.xml not found\n";
    print getTime() . " ABORT:\tProgram exit\n";
    exit;
  }
}

# ------------------------------------------------------------------------------
# Action word: InitRepo
#
# Arguments:   tokenlabel, slot, so-pin, pin
#
# Description: Initialize the repositories located in the conf.xml
# ------------------------------------------------------------------------------
sub initRepo {
  my ($row, $tokenlabel, $slot, $so_pin, $pin) = @_;
  my @rc = ();

  printActionWordHeader($row, "InitRepo()");

  @rc = runCommand("softhsm --init-token --slot $slot --label $tokenlabel --so-pin $so_pin --pin $pin");
  checkReturnCode($rc[2], 0);
}


# ------------------------------------------------------------------------------
# Action word: LoadZone
#
# Arguments:   zone policy input output
#
# Description: Load a new zone in the zone list.
#              Runs `ods-ksmutil zone add -z zone -p policy -i input -o output`.
# ------------------------------------------------------------------------------
sub loadZone {
  my ($row, $zone, $policy, $input, $output) = @_;
  my @rc = ();

  printActionWordHeader($row, "LoadZone()");

  @rc = runCommand("ods-ksmutil zone add -z " . $zone . " -p " . $policy . " -i " . $input . " -o " . $output);
  checkReturnCode($rc[2], 0);
}


# ------------------------------------------------------------------------------
# Action word: RemoveZone
#
# Arguments:   zone
#
# Description: Remove a zone in the zone list.
#              Runs `ods-ksmutil zone delete -z zone`
# ------------------------------------------------------------------------------
sub removeZone {
  my ($row, $zone) = @_;
  my @rc = ();

  printActionWordHeader($row, "RemoveZone()");

  @rc = runCommand("ods-ksmutil zone delete -z " . $zone);
  checkReturnCode($rc[2], 0);
}


# ------------------------------------------------------------------------------
# Action word: UpdateAll
#
# Arguments:
#
# Description: Updates all configs.
#              Runs `ods-ksmutil update all`.
# ------------------------------------------------------------------------------
sub updateAll {
  my ($row) = @_;
  my @rc = ();

  printActionWordHeader($row, "UpdateAll()");

  @rc = runCommand("ods-ksmutil update all");
  checkReturnCode($rc[2], 0);
}


# ------------------------------------------------------------------------------
# Action word: BackupAll
#
# Arguments:
#
# Description: Marks all repositories as backed up.
#              Runs `ods-ksmutil backup done`.
# ------------------------------------------------------------------------------
sub backupAll {
  my ($row) = @_;
  my @rc = ();

  printActionWordHeader($row, "BackupAll()");

  @rc = runCommand("ods-ksmutil backup done");
  checkReturnCode($rc[2], 0);
}


# ------------------------------------------------------------------------------
# Action word: RetireKSK
#
# Arguments:   keytag
#
# Description: Retire one of the active KSK's. There must be more then 1 active
#              key in this scenario.
#              Runs `ods-ksmutil key ksk-retire -x [keytag]`
# ------------------------------------------------------------------------------
sub retireKSK {
  my ($row, $keytag) = @_;
  my @rc = ();

  printActionWordHeader($row, "RetireKSK()");

  @rc = runCommand("ods-ksmutil key ksk-retire -x $keytag");
  checkReturnCode($rc[2], 0);
}


# ------------------------------------------------------------------------------
# Action word: DSPublished
#
# Arguments:   keytag
#
# Description: Authorize the automatic KSK rollover by telling DS is being
#              published by parent
#              Runs `ods-ksmutil ds-seen -x [keytag]`
# ------------------------------------------------------------------------------
sub DSPublished {
  my ($row, $keytag) = @_;
  my @rc = ();

  printActionWordHeader($row, "DSPublished()");

  @rc = runCommand("ods-ksmutil ds-seen -x $keytag");
  checkReturnCode($rc[2], 0);
}


# ------------------------------------------------------------------------------
# Action word: RollOverKSK
#
# Arguments:   zone, keytag
#
# Description: Initiate manual KSK rollover for given zone file
#              Runs `ods-ksmutil backup done`
#                   `ods-ksmutil key roll -z [zone] -k [keytag]`
# ------------------------------------------------------------------------------
sub rolloverKSK {
  my ($row, $zone, $keytag) = @_;
  my @rc = ();

  printActionWordHeader($row, "RolloverKSK()");

  @rc = runCommand("ods-ksmutil backup done");
  checkReturnCode($rc[2], 0);

  @rc = runCommand("ods-ksmutil key ksk-roll -z $zone -k $keytag");
  checkReturnCode($rc[2], 0);
}


# ------------------------------------------------------------------------------
# Action word: RollOverZSK
#
# Arguments:   zone
#
# Description: Manually starts the ZSK rollover for given zone file.
# ------------------------------------------------------------------------------
sub rolloverZSK {
  my ($row, $zone) = @_;

  printActionWordHeader($row, "RolloverZSK()");

  print getTime() . "Manually start the ZSK rollover for zone file $zone";
}


# ------------------------------------------------------------------------------
# Action word: CountRRSets
#
# Arguments:   input rrset expected
#
# Description: Counts number of occurences of given RRSET.
# ------------------------------------------------------------------------------
sub countRRSets {
  my ($row, $input, $rrset, $expected) = @_;
  my @rc = ();

  printActionWordHeader($row, "CountRRSets()");

  @rc = runCommand("grep $rrset $input | grep -v RRSIG | wc -l");
  checkReturnCode($rc[2], 0);
  checkReturnedOutput($rc[0], $expected);
}


# ------------------------------------------------------------------------------
# Action word: CheckKeyTag
#
# Arguments:   input zone keytype status
#
# Description: Compares keytag in signed zone with ODS key list. When the 
#              combination [zone][keytype][state] results in more then one key 
#              tag, all key tags will be selected and used to check the input 
#              zone.
#
#                ods-ksmutil key list -v 
#                ods-ksmutil key list -v | awk '{if (($1 == "nl") && ($2 == "ZSK") && ($3 == "active")) print $8}'
#                ods-ksmutil key list -v | awk '{if (($1 == "nl") && ($2 == "KSK") && ($3 == "active")) print $8}'
#                ods-ksmutil key list -v | awk '{if (($1 == "nl") && (($2 == "KSK") || ($2 == "ZSK")) && ($3 == "active")) print $8}'
# ------------------------------------------------------------------------------
sub checkKeytag {
  my ($row, $input, $zone, $keytype, $status) = @_;
  my @keytag = ();
  my @keylist = ();
  my $keylist = 0;
  my @values;

  printActionWordHeader($row, "CheckKeytag()");

  @keytag = runCommand("ods-ksmutil key list -v | awk \'{if ((\$1 == \"$zone\") && (\$2 == \"$keytype\") && (\$3 == \"$status\")) print \$8}'");
  if (!defined $keytag[0] || $keytag[0] == 0) {
    print getTime() . " FAILED:\tods-ksmutil key list -v returned no key tag\n";
    print getTime() . " ABORT:\tProgram exit\n";
    exit;
  }

  unless (-e $input) {
    print getTime() . " FAILED:\t$input not found\n";
    print getTime() . " ABORT:\tProgram exit\n";
    exit;
  }

  if (uc $keytype eq "KSK") {
    @keylist = runCommand("find_ksk_id_used_for_signing.sh $input");
    if (!defined $keylist[0] || $keylist[0] == 0) {
      print getTime() . " FAILED:\tNo KSKs used for signatures in this zone found\n";
      print getTime() . " ABORT:\tProgram exit\n";
      exit;
    }
  } elsif (uc $keytype eq "ZSK") {
    @keylist = runCommand("find_ksk_id_used_for_signing.sh $input");
    if (!defined $keylist[0] || $keylist[0] == 0) {
      print getTime() . " FAILED:\tNo ZSKs used for signatures in this zone found\n";
      print getTime() . " ABORT:\tProgram exit\n";
      exit;
    }
  } else {
    print getTime() . " ERROR:\tUnknown key type: $keytype";
    print getTime() . " ABORT:\tProgram exit\n";
    exit;
  }

  if (!defined $keylist[0] || $keylist[0] < 0) {
    print getTime() . " ERROR:\tkeylist < 0";
    print getTime() . " ABORT:\tProgram exit\n";
    exit;
  } elsif (!defined $keylist[0] || $keylist[0] == 0) {
    print getTime() . " FAILED:\tNo KSKs or ZSKs used for signatures in this zone found\n";
    print getTime() . " ABORT:\tProgram exit\n";
    exit;
  } else {
    @values = split("\n", $keytag[0]);
    foreach my $expected ($keylist[0]) {
      foreach my $found (@values) {
        if (checkForEqual(uc $found, uc $expected)) {
          print getTime() . " PASSED:\t(Check for Equal) - Expected: $expected Found: $found\n";
        } else {
          print getTime() . " FAILED:\t(Check for Equal) - Expected: $expected Found: $found\n";
        }
      }
    }
  }
}


# ------------------------------------------------------------------------------
# Action word: CheckLog
#
# Arguments:   string
#
# Description: Check whether the given string matches somewhere in the last 50
#              lines of the logging ( /var/log/messages ).
#              Runs `tail -n 50 /var/log/messages|grep [today]`.
# ------------------------------------------------------------------------------
sub checkLog {
  my ($row, $string) = @_;
  my @rc = ();

  printActionWordHeader($row, "CheckLog()");

  @rc = runCommand("tail -n 50 /var/log/messages|grep \'$string\'");
  #@rc = runCommand("tail -n 50 /var/log/messages | awk \'{print \$4}\' | grep $string");
  checkReturnCode($rc[2], 0);
  #checkReturnedOutput($rc[2], $string);

  if (!defined $rc[0]) {
    print getTime(). " FAILED:\t(Check for Like) - Expected: $string Found: \n";
    return;
  }

  if (checkForLike($rc[0], $string)) {
    print getTime() . " PASSED:\t(Check for Like) - Expected: $string Found: $rc[0]\n";
  } else {
    print getTime() . " FAILED:\t(Check for Like) - Expected: $string Found: $rc[0]\n";
  }
}


# ------------------------------------------------------------------------------
# Action word: CheckFilePresent
#
# Arguments:   file
#
# Description: Checks if the file at given path is present.
# ------------------------------------------------------------------------------
sub checkFilePresent {
  my ($row, $file) = @_;


  printActionWordHeader($row, "CheckFilePresent($file)");

  if (-e $file) {
    print getTime() . " PASSED:\t(Check for Equal) - Expected: $file Found: $file\n";
  } else {
    print getTime() . " FAILED:\t(Check for Equal) - Expected: $file Found: \n";
  }
}


# ------------------------------------------------------------------------------
# Action word: &cont
#
# Arguments:
#
# Description: Special action word which has to be present if the argument &cont
#              is found.
# ------------------------------------------------------------------------------
sub actionwordContinue {
  # print "actionwordContinue()\n";
}


# ------------------------------------------------------------------------------
# Action word: TestCase
#
# Arguments:   identifier description
#
# Description:
# ------------------------------------------------------------------------------
sub testCase {
  my ($row, $identifier, $description) = @_;

  print "\n";
  printChar('-', 79);
  print "\nTestcase - ", $identifier, ": ", $description, "\n";
}


# ------------------------------------------------------------------------------
# Action word: TestCondition
#
# Arguments:   identifier description
#
# Description:
# ------------------------------------------------------------------------------
sub testCondition {
  my ($row, $identifier, $description) = @_;

  print "Testcondition - ", $identifier, ": ", $description, "\n";
  printChar('-', 79);
  print "\n";

  $testcondition = $identifier;
}

# - End of reserved action words -----------------------------------------------


# - action word helper subroutines ---------------------------------------------

# ------------------------------------------------------------------------------
# Subroutine:  checkForEqual
#
# Arguments:   found expected
#
# Description: Check if first argument is equal to second argument.
# ------------------------------------------------------------------------------
sub checkForEqual {
  my ($found, $expected) = @_ ;

  if (defined $found && defined $expected) {
    return $found eq $expected;
  } else {
    return 0;
  }
}

# ------------------------------------------------------------------------------
# Subroutine:  checkForNotEqual
#
# Arguments:   found expected
#
# Description: Check if first argument is not equal to second argument.
# ------------------------------------------------------------------------------
sub checkForNotEqual {
  my ($found, $expected) = @_ ;

  return $found ne $expected;
}

# ------------------------------------------------------------------------------
# Subroutine:  checkForLike
#
# Arguments:   str like
#
# Description: Check if first argument is like second argument.
#              - Returns -1 if first arguments is not like second argument.
#              - Returns > -1 if first arguments is like second argument. 
# ------------------------------------------------------------------------------
sub checkForLike {
  my ($str, $like) = @_ ;

  return index($str, $like) > -1 ? 1 : 0;
}

# ------------------------------------------------------------------------------
# Subroutine:  checkForNotLike
#
# Arguments:   str like
#
# Description: Check if first argument is not like second argument.
# ------------------------------------------------------------------------------
sub checkForNotLike {
  my ($str, $like) = @_ ;

  return index($str, $like) < 0 ? 1 : 0;
}

# ------------------------------------------------------------------------------
# Subroutine:  checkReturnedOutput
#
# Arguments:
#
# Description:
# ------------------------------------------------------------------------------
sub checkReturnedOutput {
  my ($found, $expected) = @_;

  if (checkForEqual($found, $expected)) {
    print getTime() . " PASSED:\t(Check for Equal) - Expected: $expected Found: $found\n";
  } else {
    print getTime() . " FAILED:\t(Check for Equal) - Expected: $expected Found: $found\n";
  }
}

# ------------------------------------------------------------------------------
# Subroutine:  checkReturnCode
#
# Arguments:
#
# Description:
# ------------------------------------------------------------------------------
sub checkReturnCode {
  my ($found, $expected) = @_;

  if (checkForEqual($found, $expected)) {
    #
  } else {
    print getTime() . " ERROR:\treturn code = $_[0]\n"
  }
}

# ------------------------------------------------------------------------------
# Subroutine:  printRunCommandOutput
#
# Arguments:   array
#
# Description: Print return value of runcommand.
# ------------------------------------------------------------------------------
sub printRunCommandOutput() {
  my @rc = ();
  my @output = ();
  my @errors = ();

  @rc = @_;
  @output = split('\n', $rc[0]) if defined $rc[0];
  @errors = split('\n', $rc[1]) if defined $rc[1];

  # print output
  foreach my $output (@output) {
    print getTime() . " OUTPUT:\t", $output, "\n";
  }

  # print errors
  foreach my $error (@errors) {
    print getTime() . " ERROR:\t", $error, "\n";
  }
}

# ------------------------------------------------------------------------------
# Subroutine:  printArguments
#
# Arguments:   arguments array
#
# Description: Print arguments of an actionword.
# ------------------------------------------------------------------------------
sub printArguments() {
#   print join "\n", map { defined $_ ? $_ : "undef" } @_, "\n";

  foreach (@_) {
    print $_, "\n" if (defined $_);
  }
  print "\n";
}

# ------------------------------------------------------------------------------
# Subroutine:  printActionWordHeader
#
# Arguments:   row# actionword
#
# Description: Print header of an actionword, row number and name of action
#              word.
# ------------------------------------------------------------------------------
sub printActionWordHeader() {
  printf "\n%s %5.5d %s\n", getTime(), $_[0], $_[1];

  printRequirement();
}

# ------------------------------------------------------------------------------
# Subroutine:  printRequirement
#
# Arguments:   
#
# Description: Prints requirement identification if action word TestCondition is
#              found (variable $testcondition is set).
# ------------------------------------------------------------------------------
sub printRequirement {
  if (($testcondition ne "") && (length($testcondition) > 0)) {
    print getTime() . " Requirement: ", $testcondition, "\n";
  }
}

# ------------------------------------------------------------------------------
# Subroutine:  printChar
#
# Arguments: char number
#
# Description: Prints character number of times.
# ------------------------------------------------------------------------------
sub printChar {
  for (my $i = 0; $i < $_[1]; $i++) {
    print "$_[0]";
  }
  #print "\n";
}

# ------------------------------------------------------------------------------
# Subroutine:  getTime
#
# Arguments: 
#
# Description: Returns time stamp, format YYYY-MM-DD HH-MM-SS.sss (s = ms).
# ------------------------------------------------------------------------------
sub getTime() {
  my $epochseconds;
  my $microseconds;
  my $second;
  my $minute;
  my $hour;
  my $mday;
  my $mon;
  my $year;
  my $wday;
  my $yday;
  my $isdst;

  ($epochseconds, $microseconds) = gettimeofday;
  ($second,$minute,$hour,$mday,$mon,$year,$wday,$yday,$isdst)=localtime($epochseconds);

  return sprintf("%4d-%02d-%02d %02d:%02d:%02d.%.3d", $year+1900, $mon+1, $mday, $hour, $minute, $second, substr($microseconds, 0, 3));
}

# ------------------------------------------------------------------------------

1;
